/*

    Copyright (c) 2017 Motas Electronics Limited (J M HAYES)
    www.motas-synth.uk


    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and any associated documentation files
    (the "Software"), to deal in the Software without restriction,
    including without limitation the rights to use, copy, modify, merge,
    publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so,
    subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    Any person wishing to distribute modifications to the Software is
    asked to send the modifications to the original developer so that
    they can be incorporated into the canonical version.  This is,
    however, not a binding provision of this license.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR
    ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
    CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
    WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


 *
 * */

#include "readscreen.h"
#include "ui_readscreen.h"
#include <QMessageBox>
#include <QDebug>

#include <QFile>
#include <QFileDialog>

#include <QTimer>
#include "QPainter"
#include <QRect>
#include <QVector>
#include <QProcess>
#include "crc.h"
#include <QThread>
#include "QMovie"

#include"motasfirmware.h"




void MotasUtility::updatePatchUI(int row, bool dataAdjusted, bool final, int type)
{
    #define PATCH_IDENTIFIER ((quint32) 0x3561706D)
    #define PATTERN_IDENTIFIER ((quint32) 0x34F3559321)
    #define SEQUENCE_IDENTIFIER ((quint32) 0x56FE099312)
    quint32 identifierType = 0;
    quint8* buf = 0;
    quint32 itemLength = 0;
    switch (type)
    {
        case MIDI_SYSEX_PATCH_TRANSFER:
        case MIDI_SYSEX_PATCH_BANK_TRANSFER:
            patchBuffer =  bankOfPatches[row];
            buf = patchBuffer;
            identifierType = PATCH_IDENTIFIER;
            itemLength = PATCH_LENGTH_BYTES_UNPACKED;
            break;
        case MIDI_SYSEX_PATTERN_TRANSFER:
        case MIDI_SYSEX_PATTERN_BANK_TRANSFER:
            patternBuffer = bankOfPatterns[row];
            buf = patternBuffer;
            identifierType = PATTERN_IDENTIFIER;
            itemLength = PATTERN_LENGTH_BYTES_UNPACKED;
            break;
        case MIDI_SYSEX_SEQUENCE_TRANSFER:
        case MIDI_SYSEX_SEQUENCE_BANK_TRANSFER:
            sequenceBuffer = bankOfSequences[row];
            buf = sequenceBuffer;
            identifierType = SEQUENCE_IDENTIFIER;
            itemLength = SEQUENCE_LENGTH_BYTES_UNPACKED;
            break;
        default:
            return;
    }

    #define NAME_LENGTH 16
    quint32 identifier  = Utilities::read32(0, buf);
    QString itemName = " ";
    if (identifier == identifierType)
    {
        QChar n[NAME_LENGTH];
        for (int i = 0; i < NAME_LENGTH; i++)
        {
            n[i] = QChar(buf[i + 8]);
            if (n[i] < 32 || n[i] >= 0x7F)
                n[i] = 32;
        }
        // populate the name field
        itemName = QString(n, NAME_LENGTH);

        if (final)
        {
          //  ui->lineEditPatchName->setText(patchName);

     //       quint32 version  = Utilities::read32(4 + 4 + NAME_LENGTH, buf) ;
           // ui->lineEditVersion->setText(QString::number(version));
            quint32 CRC32  = Utilities::read32(4, buf);

            //ui->lineEditCRC32->setText(QString::number(CRC32));

            crcInit();
            // offset of 8 to miss the CRC and identifier bytes
            crc crcVal = crcFastSTMF4(buf + 8, itemLength - 8); // calculate CRC of all data (excluding first 8 bytes, and final 4)
            qDebug() <<  "computed CRC: " << crcVal << "for " << itemLength << " byte item";


            if (dataAdjusted)
            {
                Utilities::write32(crcVal, 4, buf); // write new CRC to the patchbuffer
                CRC32 = crcVal;
            }

            bool crcCorrect = FALSE;
            if (crcVal == CRC32)
            {
                crcCorrect = TRUE;

            } else
            {
                qDebug() <<  "CRC wrong!";
            }
            ui->labelIsCorrupted->setVisible(!crcCorrect);
            ui->pushButtonSave->setEnabled(crcCorrect);
            ui->pushButtonSend->setEnabled(crcCorrect);
         //   ui->checkBoxCRCcorrect->setChecked(crcCorrect);
        }
    } else
    {
        switch (type)
        {
            case MIDI_SYSEX_PATCH_TRANSFER:
            case MIDI_SYSEX_PATCH_BANK_TRANSFER:
                memset(buf, 0, PATCH_LENGTH_BYTES_UNPACKED);
                break;
            case MIDI_SYSEX_PATTERN_TRANSFER:
            case MIDI_SYSEX_PATTERN_BANK_TRANSFER:
                memset(buf, 0, PATTERN_LENGTH_BYTES_UNPACKED);
                break;
            case MIDI_SYSEX_SEQUENCE_TRANSFER:
            case MIDI_SYSEX_SEQUENCE_BANK_TRANSFER:
                memset(buf, 0, SEQUENCE_LENGTH_BYTES_UNPACKED);
                break;
        }

        qDebug() << "Identifier incorrect: " << QString::number(identifier, 16);
        if (final)
        {
           // ui->lineEditPatchName->setText(" ");
       //     ui->checkBoxCRCcorrect->setChecked(false);
            ui->pushButtonSave->setEnabled(!true);
            ui->pushButtonSend->setEnabled(!true);

            //ui->lineEditCRC32->setText("N/A");
          //  ui->lineEditVersion->setText(" ");
        }
    }


    QListWidgetItem* item;
    switch (type)
    {
        case MIDI_SYSEX_PATCH_TRANSFER:
        case MIDI_SYSEX_PATCH_BANK_TRANSFER:
            item = ui->listWidgetPatches->item(row);
            if (final)
            {
                ui->listWidgetPatches->setFocus();
                ui->listWidgetPatches->scrollToItem(item);            
            }
            break;
        case MIDI_SYSEX_PATTERN_TRANSFER:
        case MIDI_SYSEX_PATTERN_BANK_TRANSFER:
            item = ui->listWidgetPatterns->item(row);
            if (final)
            {
                ui->listWidgetPatterns->setFocus();
                ui->listWidgetPatterns->scrollToItem(item);
            }
            break;
        case MIDI_SYSEX_SEQUENCE_TRANSFER:
        case MIDI_SYSEX_SEQUENCE_BANK_TRANSFER:
            item = ui->listWidgetSequences->item(row);
            if (final)
            {
                ui->listWidgetSequences->setFocus();
                ui->listWidgetSequences->scrollToItem(item);
            }
            break;
            default:
            return;
    }
    // qDebug() << "row " << row;
    if (final)
        ui->lineEditName->setText(itemName);
    if (row < 9)
        item->setText(QString("[0") + QString::number(1 + row) + "] " + itemName);
    else
        item->setText(QString("[") + QString::number(1 + row) + "] " + itemName);
}



/** overwrite input buffer with converted data (from MIDI data to actual data sent of interest)
 *
 */
quint32 MotasUtility::convertRawMIDI(quint8* buf, quint32 totalSize)
{
    if (totalSize < 128)
    {
        qDebug() << "ERROR";
        return 0 ;
    }

    quint32 bufIndex = SYSEX_HEADER_BYTES;
    quint32 convertedIndex = 0;
    quint16 index = 0;
    while (bufIndex < totalSize)
    {
        // decode the MIDI buffer to the original form (with values > 0x7F)

 //       if (bufIndex < 4096* 3)
 //           qDebug() << "Buffer: " << QString::number(buf[bufIndex], 16) << " bufIndex: " << bufIndex;
        if (buf[bufIndex] == 0xF7)
        {
            // restart
            index = 0;
            bufIndex += SYSEX_HEADER_BYTES + 1;
        }
        if (bufIndex >= totalSize)
            break;
        qint16 c = Utilities::convertMIDIbytes(buf[bufIndex], index);
        if (c >= 0)
        {
            //qDebug() << "convertedI:" << convertedI;
            buf[convertedIndex++] = c;
        }
        index++;
        bufIndex++;
        //qDebug() << "Row "<< row << " bufIndex " << bufIndex << " totalSize " << totalSize << " convertedI " << convertedI;
     }

    return convertedIndex;
}


void MotasUtility::on_pushButtonRequest_clicked()
{
    switch(ui->tabWidgetDataType->currentIndex())
    {
        case 0:
            sendMIDISysExCommand("get patch cmd", MIDI_SYSEX_REQUEST_PATCH, 0);
            break;
        case 1:
            sendMIDISysExCommand("get patterns cmd", MIDI_SYSEX_REQUEST_PATTERN, 0);
            break;
        case 2:
            sendMIDISysExCommand("get sequences cmd", MIDI_SYSEX_REQUEST_SEQUENCE, 0);
            break;
        default:
            break;
    }
}


void MotasUtility::on_pushButtonSave_clicked()
{
    QString fileNameDef = fileNameLocationDefault +"/"+ ui->lineEditName->text();

    fileNameDef  = QString::fromUtf8(fileNameDef.toUtf8());

    qDebug() << fileNameDef;
    QString fileName ;
    switch(ui->tabWidgetDataType->currentIndex())
    {
        case 0:
            fileName =  QFileDialog::getSaveFileName(this,"save patch",  fileNameDef.trimmed() + "." + PATCH_FILE_EXTENSION, "*." +  PATCH_FILE_EXTENSION);
            break;
        case 1:
            fileName =  QFileDialog::getSaveFileName(this,"save pattern",  fileNameDef.trimmed() + "." +PATTERN_FILE_EXTENSION, "*." + PATTERN_FILE_EXTENSION);
            break;
        case 2:
            fileName =  QFileDialog::getSaveFileName(this,"save sequence",  fileNameDef.trimmed() + "." +SEQUENCE_FILE_EXTENSION, "*." + SEQUENCE_FILE_EXTENSION);
            break;
        default:
            break;
    }


    if (fileName.isEmpty())
        return;

    QFileInfo fileInfo(fileName);
    fileNameLocationDefault = fileInfo.absolutePath() ;


    qDebug() << fileName;
    QFile f(fileName);
    f.open(QIODevice::WriteOnly);
    switch(ui->tabWidgetDataType->currentIndex())
    {
        case 0:
            f.write((char*) this->patchBuffer, PATCH_LENGTH_BYTES_UNPACKED);
            break;
        case 1:
            f.write((char*) this->patternBuffer, PATTERN_LENGTH_BYTES_UNPACKED);
            break;
        case 2:
            f.write((char*) this->sequenceBuffer, SEQUENCE_LENGTH_BYTES_UNPACKED);
            break;
        default:
            break;
    }
    f.close();
}


void MotasUtility::on_pushButtonSend_clicked()
{

    quint8 preset = ui->comboBoxDestination->currentIndex();

    quint8 bank = 0;

    bank = 0x7E; // only send to RAM
    /*
    if (!ui->checkBoxFlashLocation->isChecked())
    {
        bank = 0x7E;
    } else
    {
        bank = ui->spinBoxBank->value();
        preset = ui->comboBoxDestination->currentIndex();
    }*/

    sendData(bank, preset);
}

void MotasUtility::on_pushButtonLoad_clicked()
{
    QString fileName;
    switch (ui->tabWidgetDataType->currentIndex())
    {
        case 0:
            fileName =  QFileDialog::getOpenFileName(this,"load patch",  fileNameLocationDefault + "/*." + PATCH_FILE_EXTENSION, "*." + PATCH_FILE_EXTENSION );

            break;
        case 1:
            fileName =  QFileDialog::getOpenFileName(this,"load pattern",  fileNameLocationDefault + "/*." + PATTERN_FILE_EXTENSION, "*." + PATTERN_FILE_EXTENSION);
            break;
        case 2:
            fileName =  QFileDialog::getOpenFileName(this,"load sequence",  fileNameLocationDefault + "/*." + SEQUENCE_FILE_EXTENSION, "*." + SEQUENCE_FILE_EXTENSION);
            break;
    }

    // read in patch from the disk



    if (fileName.isEmpty())
        return;

    QFileInfo fileInfo(fileName);
    fileNameLocationDefault = fileInfo.absolutePath() ;


    QFile f(fileName);
    qDebug() << fileName << "file size: " << f.size();

    switch (ui->tabWidgetDataType->currentIndex())
    {
        case 0:
            if (f.size() != PATCH_LENGTH_BYTES_UNPACKED)
            {
                QMessageBox::warning(this, "Error", "ERROR IN FILE SIZE");
            } else
            {
                f.open(QIODevice::ReadOnly);

                f.read((char*) this->patchBuffer, PATCH_LENGTH_BYTES_UNPACKED);
                updatePatchUI(ui->listWidgetPatches->currentRow(), false, true, MIDI_SYSEX_PATCH_TRANSFER);
            }
            break;
        case 1:
            if (f.size() != PATTERN_LENGTH_BYTES_UNPACKED)
            {
                QMessageBox::warning(this, "Error", "ERROR IN FILE SIZE");
            } else
            {
                f.open(QIODevice::ReadOnly);

                f.read((char*) this->patternBuffer, PATTERN_LENGTH_BYTES_UNPACKED);
                updatePatchUI(ui->listWidgetPatterns->currentRow(), false, true, MIDI_SYSEX_PATTERN_TRANSFER);
            }
            break;
        case 2:
            if (f.size() != SEQUENCE_LENGTH_BYTES_UNPACKED)
            {
                QMessageBox::warning(this, "Error", "ERROR IN FILE SIZE");
            } else
            {
                f.open(QIODevice::ReadOnly);

                f.read((char*) this->sequenceBuffer, SEQUENCE_LENGTH_BYTES_UNPACKED);
                updatePatchUI(ui->listWidgetSequences->currentRow(), false, true, MIDI_SYSEX_SEQUENCE_TRANSFER);
            }
            break;

    }

    f.close();
}

void MotasUtility::on_pushButtonSaveBank_clicked()
{ 
    QString fileNameDef = fileNameLocationDefault + "/bank" + QString::number(ui->comboBoxBank->currentIndex());

    switch(ui->tabWidgetDataType->currentIndex())
    {
        case 0:
            fileNameDef += "_patches";
            break;
        case 1:
            fileNameDef += "_patterns";
            break;
        case 2:
            fileNameDef += "_sequences";
            break;
        default:
            break;
    }



    fileNameDef  = QString::fromUtf8(fileNameDef.toUtf8());

    qDebug() << fileNameDef;
    QString fileName;

    switch(ui->tabWidgetDataType->currentIndex())
    {
        case 0:
            fileName =  QFileDialog::getSaveFileName(this,"save bank of patches",  fileNameDef.trimmed()+ "." +
                                                      BANK_FILE_TYPE_EXTENSION + PATCH_FILE_EXTENSION, "*." + BANK_FILE_TYPE_EXTENSION + PATCH_FILE_EXTENSION);
            break;
        case 1:
            fileName =  QFileDialog::getSaveFileName(this,"save bank of patterns",  fileNameDef.trimmed()+ "." +
                                                     BANK_FILE_TYPE_EXTENSION + PATTERN_FILE_EXTENSION, "*." + BANK_FILE_TYPE_EXTENSION + PATTERN_FILE_EXTENSION);
            break;
        case 2:
            fileName =  QFileDialog::getSaveFileName(this,"save bank of sequences",  fileNameDef.trimmed()+"." +
                                                      BANK_FILE_TYPE_EXTENSION + SEQUENCE_FILE_EXTENSION, "*." + BANK_FILE_TYPE_EXTENSION + SEQUENCE_FILE_EXTENSION);
            break;
        default:
            break;
    }


    if (fileName.isEmpty())
        return;

    QFileInfo fileInfo(fileName);
    fileNameLocationDefault = fileInfo.absolutePath() ;
    qDebug() << fileNameLocationDefault;
    qDebug() << fileName;
    QFile f(fileName);
    f.open(QIODevice::WriteOnly);


    switch(ui->tabWidgetDataType->currentIndex())
    {
        case 0:
        {
            quint8 buffer[BANK_OF_PATCHES_LENGTH * PATCH_LENGTH_BYTES_UNPACKED];
            for (int i = 0; i < BANK_OF_PATCHES_LENGTH; i++)
            {
                memcpy(&buffer[i * PATCH_LENGTH_BYTES_UNPACKED], bankOfPatches[i], PATCH_LENGTH_BYTES_UNPACKED);
            }
            f.write((char*) buffer, BANK_OF_PATCHES_LENGTH * PATCH_LENGTH_BYTES_UNPACKED);
            break;

        }
        case 1:
        {
            quint8 buffer[BANK_OF_PATTERNS_LENGTH * PATTERN_LENGTH_BYTES_UNPACKED];
            for (int i = 0; i < BANK_OF_PATTERNS_LENGTH; i++)
            {
                memcpy(&buffer[i * PATTERN_LENGTH_BYTES_UNPACKED], bankOfPatterns[i], PATTERN_LENGTH_BYTES_UNPACKED);
            }
            f.write((char*) buffer, BANK_OF_PATTERNS_LENGTH * PATTERN_LENGTH_BYTES_UNPACKED);
            break;
        }
        case 2:
        {
            quint8 buffer[BANK_OF_SEQUENCES_LENGTH * SEQUENCE_LENGTH_BYTES_UNPACKED];
            for (int i = 0; i < BANK_OF_SEQUENCES_LENGTH; i++)
            {
                memcpy(&buffer[i * SEQUENCE_LENGTH_BYTES_UNPACKED], bankOfSequences[i], SEQUENCE_LENGTH_BYTES_UNPACKED);
            }
            f.write((char*) buffer, BANK_OF_SEQUENCES_LENGTH * SEQUENCE_LENGTH_BYTES_UNPACKED);
            break;
        }
        default:
            break;
    }

    f.close();
}


void MotasUtility::on_pushButtonLoadBank_clicked()
{
    QString fileName;
    switch(ui->tabWidgetDataType->currentIndex())
    {
        case 0:
            fileName =  QFileDialog::getOpenFileName(this,"load patch bank",  fileNameLocationDefault + "/*." +
                                                      BANK_FILE_TYPE_EXTENSION + PATCH_FILE_EXTENSION, "*." + BANK_FILE_TYPE_EXTENSION + PATCH_FILE_EXTENSION);
            break;
        case 1:
            fileName =  QFileDialog::getOpenFileName(this,"load pattern bank",  fileNameLocationDefault + "/*." +
                                                       BANK_FILE_TYPE_EXTENSION + PATTERN_FILE_EXTENSION, "*." + BANK_FILE_TYPE_EXTENSION + PATTERN_FILE_EXTENSION);
            break;
        case 2:
            fileName =  QFileDialog::getOpenFileName(this,"load sequence bank",  fileNameLocationDefault + "/*." +
                                                       BANK_FILE_TYPE_EXTENSION + SEQUENCE_FILE_EXTENSION, "*." + BANK_FILE_TYPE_EXTENSION + SEQUENCE_FILE_EXTENSION);
            break;
        default:
            break;
    }
    if (fileName.isEmpty())
        return;


    QFileInfo fileInfo(fileName);
    fileNameLocationDefault = fileInfo.absolutePath() ;


    QFile f(fileName);
    qDebug() << fileName << "file size: " << f.size();

    switch(ui->tabWidgetDataType->currentIndex())
    {
        case 0:
            if (f.size() != PATCH_LENGTH_BYTES_UNPACKED * BANK_OF_PATCHES_LENGTH)
            {
                QMessageBox::warning(this, "Error", "ERROR IN FILE SIZE");
            } else
            {
                f.open(QIODevice::ReadOnly);

                for (int i = 0; i < BANK_OF_PATCHES_LENGTH; i++)
                {
                    f.read((char*) bankOfPatches[i], PATCH_LENGTH_BYTES_UNPACKED);
                    updatePatchUI(i, false, i == 0, MIDI_SYSEX_PATCH_TRANSFER);
                }
            }
            break;
        case 1:
            if (f.size() != PATTERN_LENGTH_BYTES_UNPACKED * BANK_OF_PATTERNS_LENGTH)
            {
                QMessageBox::warning(this, "Error", "ERROR IN FILE SIZE");
            } else
            {
                f.open(QIODevice::ReadOnly);

                for (int i = 0; i < BANK_OF_PATTERNS_LENGTH; i++)
                {
                    f.read((char*) bankOfPatterns[i], PATTERN_LENGTH_BYTES_UNPACKED);
                    updatePatchUI(i, false, i == 0, MIDI_SYSEX_PATTERN_TRANSFER);
                }
            }
            break;
        case 2:
            if (f.size() != SEQUENCE_LENGTH_BYTES_UNPACKED * BANK_OF_SEQUENCES_LENGTH)
            {
                QMessageBox::warning(this, "Error", "ERROR IN FILE SIZE");
            } else
            {
                f.open(QIODevice::ReadOnly);

                for (int i = 0; i < BANK_OF_SEQUENCES_LENGTH; i++)
                {
                    f.read((char*) bankOfSequences[i], SEQUENCE_LENGTH_BYTES_UNPACKED);
                    updatePatchUI(i, false, i == 0, MIDI_SYSEX_SEQUENCE_TRANSFER);
                }
            }
            break;
        default:
            break;
    }
    f.close();
}

void MotasUtility::on_pushButtonRequestBank_clicked()
{

    switch (ui->tabWidgetDataType->currentIndex())
    {
        case 0:
            sendMIDISysExCommand("get bank patches cmd", MIDI_SYSEX_REQUEST_PATCHES_BANK, ui->comboBoxBank->currentIndex() );
            break;
        case 1:
            sendMIDISysExCommand("get patterns bank cmd", MIDI_SYSEX_REQUEST_PATTERN_BANK, ui->comboBoxBank->currentIndex() );
            break;
        case 2:
            sendMIDISysExCommand("get sequence bank cmd", MIDI_SYSEX_REQUEST_SEQUENCE_BANK, ui->comboBoxBank->currentIndex() );
            break;
        default:
            break;
    }
}


void MotasUtility::clearAll(int type)
{

    switch (type)
    {
        case 0:
            for (int i = 0; i < bankOfPatches.length(); i++)
             {
                  quint8* b = bankOfPatches.at(i);
                  memset(b, 0, PATCH_LENGTH_BYTES_UNPACKED); // 'erase' patch

                  QListWidgetItem* item = ui->listWidgetPatches->item(i);
                  if (i < 9)
                      item->setText(QString("[0") + QString::number(1 + i) + "]  ");
                  else
                      item->setText(QString("[") + QString::number(1 + i) + "]  ");

             }
            ui->listWidgetPatches->setCurrentRow(0);
            break;
        case 1:
            for (int i = 0; i < bankOfPatterns.length(); i++)
            {
                quint8* b = bankOfPatterns.at(i);
                memset(b, 0, PATTERN_LENGTH_BYTES_UNPACKED); // 'erase' patch

                QListWidgetItem* item = ui->listWidgetPatterns->item(i);
                if (i < 9)
                    item->setText(QString("[0") + QString::number(1 + i) + "]  ");
                else
                    item->setText(QString("[") + QString::number(1 + i) + "]  ");

            }
            ui->listWidgetPatterns->setCurrentRow(0);
            break;
        case 2:
            for (int i = 0; i < bankOfSequences.length(); i++)
            {
                quint8* b = bankOfSequences.at(i);
                memset(b, 0, SEQUENCE_LENGTH_BYTES_UNPACKED); // 'erase' patch

                QListWidgetItem* item = ui->listWidgetSequences->item(i);
                if (i < 9)
                    item->setText(QString("[0") + QString::number(1 + i) + "]  ");
                else
                    item->setText(QString("[") + QString::number(1 + i) + "]  ");

            }
            ui->listWidgetSequences->setCurrentRow(0);
            break;
    }
}

void MotasUtility::on_pushButtonClearData_clicked()
{
    clearAll(ui->tabWidgetDataType->currentIndex());
    ui->lineEditName->clear();
 //   ui->lineEditVersion->clear();
  //  ui->checkBoxCRCcorrect->setChecked(false);
}


void MotasUtility::on_listWidgetPatterns_currentRowChanged(int currentRow)
{
     updatePatchUI(currentRow, false, true, MIDI_SYSEX_PATTERN_TRANSFER);
}
void MotasUtility::on_listWidgetSequences_currentRowChanged(int currentRow)
{
    updatePatchUI(currentRow, false, true, MIDI_SYSEX_SEQUENCE_TRANSFER);
}



void MotasUtility::on_listWidgetPatches_currentRowChanged(int currentRow)
{
     updatePatchUI(currentRow, false, true, MIDI_SYSEX_PATCH_TRANSFER);
}

void MotasUtility::on_lineEditName_editingFinished()
{
    qDebug() << " name edited";
    QByteArray b = QByteArray(ui->lineEditName->text().toUtf8(), ui->lineEditName->text().length());

    switch (ui->tabWidgetDataType->currentIndex())
    {
        case 0:
            memset(&patchBuffer[8], 32, NAME_LENGTH);
            break;
        case 1:
            memset(&patternBuffer[8], 32, NAME_LENGTH);
            break;
        case 2:
            memset(&sequenceBuffer[8], 32, NAME_LENGTH);
            break;
    }

    for (int i = 0; i < ui->lineEditName->text().length() ; i++)
    {
        quint8 val = b.at(i);
        if (val < 32 || val >= 0x7F)
            val = 32;
        switch (ui->tabWidgetDataType->currentIndex())
        {
            case 0:
                patchBuffer[i + 8] = val;
                break;
            case 1:
                patternBuffer[i + 8] = val;
                break;
            case 2:
                sequenceBuffer[i + 8] = val;
               break;
        }
    }

    updateCurrentDisplayData();
}

void MotasUtility::on_pushButtonSendBank_clicked()
{
    // send all patches to Motas, to be stored in FLASH
    quint32 bytesToSend = 0;
    quint8* outBuffer = 0;

    quint8 model = ui->spinBoxMotasModel->value();
    quint8 channel = ui->spinBoxChannel->value() - 1;
    switch (ui->tabWidgetDataType->currentIndex())
    {
        case 0:
        {
            outBuffer = new quint8[PATCH_LENGTH_BYTES_UNPACKED * BANK_OF_PATCHES_LENGTH *2];
            quint8 buffer[PATCH_LENGTH_BYTES_UNPACKED * BANK_OF_PATCHES_LENGTH];

            for (int i = 0; i < BANK_OF_PATCHES_LENGTH; i++)
            {
                memcpy(&buffer[i * PATCH_LENGTH_BYTES_UNPACKED], bankOfPatches[i], PATCH_LENGTH_BYTES_UNPACKED);
            }
            bytesToSend
                    = motasFirmware->prepareSysExPacketsFromUnpacked(outBuffer,buffer,
                    PATCH_LENGTH_BYTES_UNPACKED * BANK_OF_PATCHES_LENGTH,
                    MIDI_SYSEX_PATCH_BANK_TRANSFER, ui->comboBoxBank->currentIndex(), 0, 0, model, channel);

        }
            break;
        case 1:
        {
            outBuffer = new quint8[PATTERN_LENGTH_BYTES_UNPACKED * BANK_OF_PATTERNS_LENGTH *2];
            quint8 buffer[PATTERN_LENGTH_BYTES_UNPACKED * BANK_OF_PATTERNS_LENGTH];

            for (int i = 0; i < BANK_OF_PATTERNS_LENGTH; i++)
            {
                memcpy(&buffer[i * PATTERN_LENGTH_BYTES_UNPACKED], bankOfPatterns[i], PATTERN_LENGTH_BYTES_UNPACKED);
            }
            bytesToSend
                    = motasFirmware->prepareSysExPacketsFromUnpacked(outBuffer,buffer,
                    PATTERN_LENGTH_BYTES_UNPACKED * BANK_OF_PATTERNS_LENGTH,
                    MIDI_SYSEX_PATTERN_BANK_TRANSFER, ui->comboBoxBank->currentIndex(), 0, 0, model, channel);

        }
            break;
        case 2:
        {
            outBuffer = new quint8[SEQUENCE_LENGTH_BYTES_UNPACKED * BANK_OF_SEQUENCES_LENGTH *2];
            quint8 buffer[SEQUENCE_LENGTH_BYTES_UNPACKED * BANK_OF_SEQUENCES_LENGTH];

            for (int i = 0; i < BANK_OF_SEQUENCES_LENGTH; i++)
            {
                memcpy(&buffer[i * SEQUENCE_LENGTH_BYTES_UNPACKED], bankOfSequences[i], SEQUENCE_LENGTH_BYTES_UNPACKED);
            }
            bytesToSend
                    = motasFirmware->prepareSysExPacketsFromUnpacked(outBuffer,buffer,
                    SEQUENCE_LENGTH_BYTES_UNPACKED * BANK_OF_SEQUENCES_LENGTH,
                    MIDI_SYSEX_SEQUENCE_BANK_TRANSFER, ui->comboBoxBank->currentIndex(), 0, 0, model, channel);

        }
        break;
    }

    this->sendDataOverMIDIDirect(outBuffer, bytesToSend);

    delete outBuffer;
    outBuffer = 0;
    //qDebug() << "bytes to send" << bytesToSend;





}

void MotasUtility::sendData(quint8 bank, quint8 preset)
{
    quint32 bytesToSend;
    quint8 model = ui->spinBoxMotasModel->value();
    quint8 channel = ui->spinBoxChannel->value() - 1;

    switch(ui->tabWidgetDataType->currentIndex())
    {
        case 0:
            bytesToSend = motasFirmware->prepareSysExPacketsFromUnpacked(rawTransferBuffer,
                      patchBuffer, PATCH_LENGTH_BYTES_UNPACKED, MIDI_SYSEX_PATCH_TRANSFER, bank, preset, 0, model, channel);

            break;
        case 1:
            bytesToSend = motasFirmware->prepareSysExPacketsFromUnpacked(rawTransferBuffer,
                  patternBuffer, PATTERN_LENGTH_BYTES_UNPACKED, MIDI_SYSEX_PATTERN_TRANSFER, bank, preset, 0, model, channel);
            break;
        case 2:
            bytesToSend = motasFirmware->prepareSysExPacketsFromUnpacked(rawTransferBuffer,
              sequenceBuffer, SEQUENCE_LENGTH_BYTES_UNPACKED, MIDI_SYSEX_SEQUENCE_TRANSFER, bank, preset, 0, model, channel);

         break;
    }


    qDebug() << "bytes to send" << bytesToSend;

    this->sendDataOverMIDIDirect(rawTransferBuffer, bytesToSend);
}


void MotasUtility::processMIDIdata(quint8 type,  QVector<quint8>* midiStreamData)
{

    ui->pushButtonSave->setEnabled(false);
    ui->pushButtonSend->setEnabled(false);


//    qDebug() << "on final";
    // progressMovie->stop();
   // int bank = this->midiStreamPatch->at(6);
//    quint8 row = this->midiStreamPatch->at(7);
//    if (row >= BANK_OF_PATCHES_LENGTH)

    quint32 len = midiStreamData->size();

    quint8 buffer[len];
    for (quint32 i = 0; i < len; i++)
    {
        buffer[i] = midiStreamData->at(i);
    }
    midiStreamData->clear();

  //  qDebug() << "msg0: " << buffer[SYSEX_HEADER_BYTES - 3] << "msg1: " << buffer[SYSEX_HEADER_BYTES - 2] << "msg2: " << buffer[SYSEX_HEADER_BYTES - 1] ;
    quint32 convertedLength = convertRawMIDI(buffer, len);

    quint32 expectedLength = 0;
    bool fullBankTransfer = false;
   // quint32 sectorId = 0;
    switch (type)
    {
        case MIDI_SYSEX_BITMAP_TRANSFER:
            expectedLength = 4096;
            break;
        case MIDI_SYSEX_OSCILLOSCOPE_TRANSFER:
            expectedLength = OSCILLOSCOPE_TRACE_NUM_POINTS*2;
            break;
        case MIDI_SYSEX_REQUEST_BULK_FLASH_DATA_SEND:

          //  sectorId =  buffer[SYSEX_HEADER_BYTES - 3] * 128 + buffer[SYSEX_HEADER_BYTES - 2];
            expectedLength = MASS_FLASH_TRANSFER_LENGTH_BYTES;
            break;
        case MIDI_SYSEX_PATCH_TRANSFER:
            expectedLength = PATCH_LENGTH_BYTES_UNPACKED;
            break;
        case MIDI_SYSEX_PATCH_BANK_TRANSFER:
            fullBankTransfer = true;
            expectedLength  = BANK_OF_PATCHES_LENGTH * PATCH_LENGTH_BYTES_UNPACKED;
            break;
        case MIDI_SYSEX_PATTERN_TRANSFER:
            expectedLength  = PATTERN_LENGTH_BYTES_UNPACKED;
            break;
        case MIDI_SYSEX_PATTERN_BANK_TRANSFER:
            fullBankTransfer = true;
            expectedLength  = BANK_OF_PATTERNS_LENGTH * PATTERN_LENGTH_BYTES_UNPACKED;
            break;
        case MIDI_SYSEX_SEQUENCE_TRANSFER:
            expectedLength  =  SEQUENCE_LENGTH_BYTES_UNPACKED;
            break;
        case MIDI_SYSEX_SEQUENCE_BANK_TRANSFER:
            fullBankTransfer = true;
            expectedLength  = BANK_OF_SEQUENCES_LENGTH * SEQUENCE_LENGTH_BYTES_UNPACKED;
            break;

        case MIDI_SYSEX_GLOBALS_TRANSFER:
            expectedLength = SETUP_DATA_LENGTH_BYTES;
            break;
        default:
            qDebug() << "unknown type!";
            QMessageBox::warning(this, "Error", "Unknown data type");
            ui->checkBoxSendBitmap->setChecked(false);
            return;
            break;
    }
    expectedLength += 4;// add on the CRC bytes

    if (convertedLength != expectedLength)
    {
        qDebug() <<  "incorrect transfer length: " << convertedLength << "bytes converted, expected: " <<  expectedLength;
        QMessageBox::warning(this, "Error", "Data length wrong");
        ui->checkBoxSendBitmap->setChecked(false);
        return;
    }
    // check to final CRC

    // final one

    // calculate CRC
    quint32 CRC32  = Utilities::read32(convertedLength - 4, buffer);

    crcInit();
    // offset of 8 to miss the CRC and identifier bytes
    crc crcVal = crcFastSTMF4(buffer, convertedLength - 4); // calculate CRC of all data (excluding first 8 bytes, and final 4)
//      qDebug() <<  "computed CRC for transfer: " << crcVal;


    QString fileNameDef = fileNameLocationDefault +"/";


    if (CRC32 == crcVal)
    {
        quint8 row = 0;
        quint32 end = 1;
        switch (type)
        {
            case MIDI_SYSEX_BITMAP_TRANSFER:
                updateMIDIreceivedScreenGrab(buffer);
                break;
            case MIDI_SYSEX_OSCILLOSCOPE_TRANSFER:
                updateOscilloscope(buffer);
                break;
            case MIDI_SYSEX_REQUEST_BULK_FLASH_DATA_SEND:
            {
                // write to disk file

                if (TRUE)
                {
                    QString fileName =  QFileDialog::getSaveFileName(this,"save FLASH memory from Motas to file",  fileNameDef + "bulk.dat");

                    if (fileName.isEmpty())
                        return;

                    QFileInfo fileInfo(fileName);
                    fileNameLocationDefault = fileInfo.absolutePath() ;




                    qDebug() << fileName;
                    QFile f(fileName);
                    f.open(QIODevice::ReadWrite);
                    f.write((char*) buffer, convertedLength - 4); // don't write the CRC at the end

                    f.close();
                }

            }
                break;
            case MIDI_SYSEX_GLOBALS_TRANSFER:
            {
                // write to disk file

                if (TRUE)
                {
                    QString fileName =  QFileDialog::getSaveFileName(this,"save setup data from Motas to file", fileNameDef +  "setupData.setup");

                    if (fileName.isEmpty())
                        return;
                    QFileInfo fileInfo(fileName);
                    fileNameLocationDefault = fileInfo.absolutePath() ;
                    qDebug() << fileName;
                    QFile f(fileName);
                    f.open(QIODevice::ReadWrite);
                    f.write((char*) buffer, convertedLength - 4); // don't write the CRC at the end

                    f.close();
                }

            }
            break;
            case MIDI_SYSEX_PATCH_TRANSFER:
            case MIDI_SYSEX_PATCH_BANK_TRANSFER:
            {
                if (fullBankTransfer)
                {
                    row = 0;
                    end = BANK_OF_PATCHES_LENGTH;
                } else
                {
                     row = ui->listWidgetPatches->currentRow();
                }
                for (quint32 i = 0; i < end; i++)
                {
                    memcpy(bankOfPatches[row], &buffer[i * PATCH_LENGTH_BYTES_UNPACKED], PATCH_LENGTH_BYTES_UNPACKED);

                    updatePatchUI(row, false, i == end - 1, MIDI_SYSEX_PATCH_TRANSFER);
                    row++;
                }
                break;
            }
            case MIDI_SYSEX_PATTERN_TRANSFER:
            case MIDI_SYSEX_PATTERN_BANK_TRANSFER:
                if (fullBankTransfer)
                {
                    row = 0;
                    end = BANK_OF_PATTERNS_LENGTH;
                } else
                {
                     row = ui->listWidgetPatterns->currentRow();
                }
                for (quint32 i = 0; i < end; i++)
                {
                    memcpy(bankOfPatterns[row], &buffer[i * PATTERN_LENGTH_BYTES_UNPACKED], PATTERN_LENGTH_BYTES_UNPACKED);

                    updatePatchUI(row, false, i == end - 1, MIDI_SYSEX_PATTERN_TRANSFER);

                    row++;
                }
                break;
            case MIDI_SYSEX_SEQUENCE_TRANSFER:
            case MIDI_SYSEX_SEQUENCE_BANK_TRANSFER:
                if (fullBankTransfer)
                {
                    row = 0;
                    end = BANK_OF_SEQUENCES_LENGTH;
                } else
                {
                     row = ui->listWidgetSequences->currentRow();
                }
                for (quint32 i = 0; i < end; i++)
                {
                    memcpy(bankOfSequences[row], &buffer[i * SEQUENCE_LENGTH_BYTES_UNPACKED], SEQUENCE_LENGTH_BYTES_UNPACKED);

                    updatePatchUI(row, false, i == end - 1, MIDI_SYSEX_SEQUENCE_TRANSFER);

                    row++;
                }
                break;
            default:
                break;
        }
    } else
    {
         QMessageBox::warning(this, "Error", "CRC incorrect - data not received correctly");
         qDebug() <<  "CRC wrong: " << crcVal << ": " << CRC32;
    }
}
void MotasUtility::updateMIDIReceiveByte(quint8 data)
{



    bool final = FALSE;
    static qint32 indexF0 = -1;

    quint8 type = 0;


    if (data == 0xF0) // start packet seen
        indexF0 = numBytesReceived;

    this->writeMidiStreamData->append(data);
    numBytesReceived++;

    if (data == 0xF7)
    {
       // qDebug() << "MIDI sysex packet received";

         if (indexF0 >= 0 && (numBytesReceived > (quint32) (indexF0 + 8)) )
        {
            if (this->writeMidiStreamData->at(indexF0 + 8) ==  MIDI_SYSEX_END_TRANSFER)
            {
                final = TRUE;
                type = this->writeMidiStreamData->at(indexF0  + 7);
            }

            bool abortEarly = FALSE;
            for (int i = 0; i < 7; i++)
            {
                int val = this->writeMidiStreamData->at(indexF0 + i);
                switch (i)
                {
                    case 0:
                        if (val != 0xF0)
                            abortEarly = TRUE;
                        break;
                    case 1:
                        if (val != MIDI_stuff::BYTE_SYSEX_ID1)
                            abortEarly= TRUE;
                        break;
                    case 2:
                        if (val != MIDI_stuff::BYTE_SYSEX_ID2)
                            abortEarly = TRUE;
                        break;
                    case 3:
                        if (val != MIDI_stuff::BYTE_SYSEX_ID3 )
                            abortEarly = TRUE;
                        break;
                    case 4:
                        if (val != ui->spinBoxMotasModel->value())
                            abortEarly = TRUE;
                        break;
                    case 5:
                        // channel, ignored.
                       // if (val != MIDI_stuff::BYTE_SYSEX_PRODUCT_ID_2 )
                       //    abortEarly = TRUE;
                        break;
                    case 6:

                        break;
                    default:
                        break;
                }
            }
            if (abortEarly)
            {
                qDebug() << " aborted - SYSex received not correctly from Motas-5";
                for (int i = 0; i < 4; i++)
                {
                    int val = this->writeMidiStreamData->at(indexF0 + i);
                    qDebug() << "byte "  << i << "value: " << val;
                }
                final = TRUE;
            } else
            {


            }

        }
        indexF0 = -1;
    }

    if (final && this->MIDItransferState == 0)
    {

        if (this->writeMidiStreamData == this->midiStreamData1)
        {
            this->writeMidiStreamData = this->midiStreamData2;
            this->readMidiStreamData = this->midiStreamData1;
        }
        else
        {
            this->writeMidiStreamData = this->midiStreamData1;
            this->readMidiStreamData = this->midiStreamData2;
        }

        //qDebug() << " data ok, bytes: " << numBytesReceived;
        MIDItransferType = type;
        this->MIDItransferState = 1;
        numBytesReceived = 0;
    }
}



void MotasUtility::on_listWidgetPatches_itemClicked(QListWidgetItem *item)
{
     Q_UNUSED(item);
    if (ui->checkBoxAutoSendData->isChecked())
    {
        sendData(0x7E,  ui->comboBoxDestination->currentIndex());
    }
}

void MotasUtility::updateCurrentDisplayData()
{
    ui->comboBoxBank->clear();
    ui->comboBoxDestination->clear();
    switch (ui->tabWidgetDataType->currentIndex())
    {
        case 0:
            for (int i = 0; i < 10; i++)
            {
                ui->comboBoxBank->addItem("FLASH bank: "+ QString::number(i ));
            }
            updatePatchUI(ui->listWidgetPatches->currentRow(), true, true, MIDI_SYSEX_PATCH_TRANSFER);
            for (int i = 0; i < 5; i++)
            {
                ui->comboBoxDestination->addItem("Preset RAM: " + QString::number(i + 1));
            }
            break;
        case 1:
            for (int i = 0; i < 7; i++)
            {
                ui->comboBoxBank->addItem("FLASH bank: "+  QString::number(i ));
            }
            updatePatchUI(ui->listWidgetPatterns->currentRow(), true, true, MIDI_SYSEX_PATTERN_TRANSFER);

            for (int i = 0; i < 15; i++)
            {
                ui->comboBoxDestination->addItem("Location: "  + QString('A' + i));
            }

            break;
        case 2:
            for (int i = 0; i < 2; i++)
            {
                ui->comboBoxBank->addItem("FLASH bank: "+  QString::number(i ));
            }

            updatePatchUI(ui->listWidgetSequences->currentRow(), true, true, MIDI_SYSEX_SEQUENCE_TRANSFER);

            for (int i = 0; i < 1; i++)
            {
                ui->comboBoxDestination->addItem("Current RAM: " + QString::number(i + 1));
            }

            break;
    }
}

void MotasUtility::on_tabWidgetDataType_currentChanged(int index)
{
    Q_UNUSED(index);
    updateCurrentDisplayData();
}



