/*

    Copyright (c) 2017 Motas Electronics Limited (J M HAYES)
    www.motas-synth.uk


    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and any associated documentation files
    (the "Software"), to deal in the Software without restriction,
    including without limitation the rights to use, copy, modify, merge,
    publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so,
    subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    Any person wishing to distribute modifications to the Software is
    asked to send the modifications to the original developer so that
    they can be incorporated into the canonical version.  This is,
    however, not a binding provision of this license.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR
    ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
    CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
    WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


 *
 * */

#ifndef READSCREEN_H
#define READSCREEN_H

#include <QMainWindow>
#include "RtMidi/RtMidi.h"

#include <QtGlobal>
#include <QDialog>
#include <QButtonGroup>
#include "utilities.h"
#include "ui_readscreen.h"
#include <QGraphicsScene>
#include "oscilloscopescreen.h"
#include "screencapture.h"
class MotasFirmware;
#include "QVector"

#define OSCILLOSCOPE_TRACE_NUM_POINTS 128
#define OSCILLOSCOPE_TRACE_X_SCALING 8
#define OSCILLOSCOPE_TRACE_Y_SCALING 6

#define NEW_FIRMWARE_START_ADDRESS 768
#define MASS_FLASH_TRANSFER_LENGTH_BYTES (NEW_FIRMWARE_START_ADDRESS * 4096L)
//#define MASS_FLASH_TRANSFER_LENGTH_BYTES (16 * 4096L) // use this in testing to transfer 16 sectors from firmware area
#define SETUP_DATA_LENGTH_BYTES 256

#define MOTAS_FIRMARE_FILE_EXTENSION QString("mbin")

#define PATCH_FILE_EXTENSION QString("patch")
#define PATTERN_FILE_EXTENSION QString("patt")
#define SEQUENCE_FILE_EXTENSION QString("seq")
#define BANK_FILE_TYPE_EXTENSION QString("b")



/* VERSION information
 *   first released version 12th October 2017  v1.0
 *
 *
 *
 *
 * */

#define VERSION "v1.0 October 2017"

namespace Ui {
   // class readScreen;
    class MotasUtility;
}

class MotasUtility : public QMainWindow {
    Q_OBJECT

private:
    QGraphicsScene* oscilloscopeScene;
     RtMidiIn *midiIn  ;
     RtMidiOut *midiOut  ;
      bool MIDIReady;
      MotasFirmware* motasFirmware;
      Utilities* utilities;
       quint32 read32(int offset, quint8 *buffer);

       void write32(quint32 val, int offset, quint8 *buffer);


       QList<quint8*> bankOfPatches;
       QList<quint8*> bankOfPatterns;
        QList<quint8*> bankOfSequences;
       quint8* rawTransferBuffer;
       quint8* patchBuffer;

       quint8* patternBuffer;
       quint8* sequenceBuffer;
    bool reportValues ;

    void resendMIDIcommands();


    //void mycallback( double deltatime, std::vector< unsigned char > *message, void *userData );
public:
    MotasUtility(QWidget *parent = 0);
    ~MotasUtility();


   void setupScreenShotView(ScreenCapture* c);

    void setUpScopeView(OscilloscopeScreen* o);


   static  void mycallback(double deltatime, std::vector< unsigned char > *message, void *userData );


   static void midiInErrorCallback(RtMidiError::Type type, const std::string &errorText, void *userData);
   static void midiOutErrorCallback(RtMidiError::Type type, const std::string &errorText, void *userData);
  // QVector<quint8>* midiStreamScreenGrab;
   QVector<quint8>* midiStreamData1;
   QVector<quint8>* midiStreamData2;

   QVector<quint8>* readMidiStreamData;
   QVector<quint8>* writeMidiStreamData;


   void updateCurrentDisplayData();

  // QVector<int>* MIDIbyteCount;
   int MIDIbyteCounter;
   void updateMIDIreceivedScreenGrab(quint8 *buffer);
   void updateOscilloscope(quint8 *buffer);
   void updateMIDIReceiveByte(quint8 data);

   quint16 countDownTimerMIDIIn;
   quint16 countDownTimerMIDIOut;
    quint32 convertRawMIDI(quint8* buf, quint32 totalSize);



    bool isConnected;

   QString fileNameLocationDefault;

   quint32 numBytesReceived;
   int MIDItransferState;
   quint8 MIDItransferType;

Ui::MotasUtility *ui;
protected:
    void changeEvent(QEvent *e);


private:

    OscilloscopeScreen *oscilloscopeCapture;
    ScreenCapture* screenCapture;

    QGraphicsPathItem* oscilloscopeLine;
    void closePorts();
    void updateMIDI(bool restart);
    QTimer *timer;

    QPixmap* pixmap;
    QImage* smallImage;
    QImage image1;
    QImage image2;
    QMovie *progressMovie ;
     QMovie *progressMovieMIDIOut ;
    int rescaleFactor;
    QImage *imageIt(quint8 *buffer);


    bool toggleLockImage;
    bool displayedImage;
    bool toggleLockOscilloscope;
    bool displayedOscilloscopeTrace;

    quint16 oscilloscopeTrace[OSCILLOSCOPE_TRACE_NUM_POINTS];
    QRect  rect;
    void calcRect();
    void generateSvgIcons();
    void memcpyReverse32(char* out, char* in);
    void memcpyForward32(char* out, char* in);
    void sendMIDISysExCommand(QString name, char msg1, char msg2);
    void test_encrypt_cbc();
    void generateFirmware(QString fileName);
    QVector<quint8> prepareSysExPacket(int j, char* buffer, quint32 packetSize, quint32 fileLength, int32_t lastPacket);
    void sendDataOverMIDI(char *buffer, quint32 fileLength, int type, quint8 msg0, quint8 msg1, quint8 msg2);
    quint32 convertForMIDISending(char* outBuf, char* inBuf, int length, quint8 type);
    int sendDataOverMIDIDirect(unsigned char* buffer, quint32 fileLength);
    //quint32 rawPatchTransferBufferLength;
    void updatePatchUI(int row, bool dataAdjusted, bool final, int state);
    void sendData(quint8 bank, quint8 preset);
    void updatePatternsUI(int row, bool dataAdjusted, bool final);

    void processMIDIdata(quint8 type, QVector<quint8> *midiStreamData);
    void clearAll(int type);
private slots:

    void on_MIDIopen_pushButton_clicked();
    void update();

    void on_pushButtonSendFirmware_clicked();


    void on_pushButtonRequestBank_clicked();

    void on_pushButtonClearData_clicked();

    void on_listWidgetPatches_currentRowChanged(int currentRow);
    void on_lineEditName_editingFinished();
    void on_pushButtonSendBank_clicked();
    void on_listWidgetPatches_itemClicked(QListWidgetItem *item);
    void on_checkBoxSendBitmap_clicked();
    void on_pushButtonSaveBank_clicked();
    void on_pushButtonLoadBank_clicked();

    void on_pushButtonRequest_clicked();
    void on_pushButtonSave_clicked();
    void on_pushButtonLoad_clicked();
    void on_pushButtonSend_clicked();
    void on_listWidgetPatterns_currentRowChanged(int currentRow);
    void on_tabWidgetDataType_currentChanged(int index);

    void on_pushButtonFindDevices_clicked();
    void on_pushButtonAbort_clicked();
    void on_comboBoxOutMIDI_currentTextChanged(const QString &arg1);
    void on_listWidgetSequences_currentRowChanged(int currentRow);


    void on_checkBoxRequestOscilloscopeData_clicked();
    void on_pushButton_clicked();
    void on_pushButtonSendBulk_clicked();

    void on_checkBoxSmoothImage_clicked();
    void on_pushButtonRequestSetup_clicked();
    void on_pushButtonSendSetup_clicked();
    void on_TabWidget_currentChanged(int index);
    void on_tabWidgetSingleOrBank_currentChanged(int index);

};




enum
{
    MIDI_SYSEX_NOT_SET  = 0,
    MIDI_SYSEX_REQUEST_BITMAP = 10,
    MIDI_SYSEX_REQUEST_OSCILLOSCOPE_TRACE = 11,
    MIDI_SYSEX_REQUEST_BULK_FLASH_DATA_SEND = 12,
    MIDI_SYSEX_RECEIVE_BULK_FLASH_DATA = 13,

    MIDI_SYSEX_REQUEST_GLOBALS = 15,

    MIDI_SYSEX_REQUEST_PATCH = 20,
    MIDI_SYSEX_REQUEST_PATCHES_BANK,

    MIDI_SYSEX_REQUEST_PATTERN = 25,
    MIDI_SYSEX_REQUEST_PATTERN_BANK,

    MIDI_SYSEX_REQUEST_SEQUENCE = 30,
    MIDI_SYSEX_REQUEST_SEQUENCE_BANK,


    MIDI_SYSEX_BITMAP_TRANSFER = 40,
    MIDI_SYSEX_OSCILLOSCOPE_TRANSFER = 41,
    MIDI_SYSEX_GLOBALS_TRANSFER = 45,

    MIDI_SYSEX_PATCH_TRANSFER = 50,
    MIDI_SYSEX_PATCH_BANK_TRANSFER,
    MIDI_SYSEX_PATTERN_TRANSFER = 55,
    MIDI_SYSEX_PATTERN_BANK_TRANSFER,
    MIDI_SYSEX_SEQUENCE_TRANSFER = 60,
    MIDI_SYSEX_SEQUENCE_BANK_TRANSFER,

    MIDI_SYSEX_FIRMWARE = 70, // 0x32
    MIDI_SYSEX_START_TRANSFER,
    MIDI_SYSEX_CONTINUE_TRANSFER,
    MIDI_SYSEX_END_TRANSFER, //0x35
    MIDI_SYSEX_START_AND_END_TRANSFER, // used for small transfers
    MIDI_SYSEX_ABORT_TRANSFER,
};// enumMIDI_SYSEX;




#define SYSEX_HEADER_BYTES  12
#define BANK_OF_PATCHES_LENGTH 50
#define BANK_OF_PATTERNS_LENGTH 8
#define BANK_OF_SEQUENCES_LENGTH 8


#define PATCH_LENGTH_BYTES_UNPACKED 4096
#define PATTERN_LENGTH_BYTES_UNPACKED 2048 //192 // using sizeof
#define SEQUENCE_LENGTH_BYTES_UNPACKED (4096L * (5 + 8))

#define LARGEST_BUFFER_LENGTH (SEQUENCE_LENGTH_BYTES_UNPACKED * 2)



#define LENGTH_OF_PATCH_NAME 16

#endif // READSCREEN_H
